#pragma once
#define BUILD_BUG_ON(condition) ((void)sizeof(char[1 - 2*!!(condition)]))
#include <cstddef> //Standard Definitions: ptrdiff_t, size_t, NULL
#include "string.h"

///NOTICE: Manual basic vector implementation saves tons of DLL file space!!!
/// Not meant to be used for bulk data processing; for API related use only.
template <typename T>
struct Vx
{
    size_t sz;
    unsigned char data_[4096-sizeof(size_t)];
    T& operator[](int index){return ((T*)data_)[index];}
    size_t size(){return sz;}
    Vx(void* beg, void* end){
        sz = (((size_t)end - (size_t)beg) / sizeof(T));
        for(size_t i = 0; i < sz; i++){(*this)[i] = ((T*)beg)[i];}
    }
    Vx(){}
};

///NOTICE: Manual basic string implementation saves tons of DLL file space!!!
/// Not meant to be used for bulk data processing; for API related use only.
struct TXT
{
    char data_[4096];
    char& operator[](int index){return data_[index];}
    size_t size(){return strlen(data_);}
    TXT(const char* data){strncpy(data_, data, 4095); data_[4095] = '\0';}
    TXT(){data_[0] = '\0';}
};

typedef unsigned long long          uint64_t;
typedef unsigned int                uint32_t;
typedef unsigned short int          uint16_t;
typedef unsigned char                uint8_t;
typedef long long                    int64_t;
typedef signed int                   int32_t;
typedef signed short int             int16_t;
typedef signed char                   int8_t;

/// (sizeof(char) == 1) ALWAYS.
/// C spec 6.5.3.4 says "When applied to an operand that has type char, 
/// unsigned char, or signed char, (or a qualified version thereof) the result is 1."
/// This guarantee is used to ensure binary compatibility in some of DELTA's code,
/// especially the object DLL interfaces and APIs.

typedef    uint64_t  U64;
typedef    uint32_t  U32;
typedef    uint16_t  U16;
typedef     uint8_t  U08;
typedef     int64_t  S64;
typedef     int32_t  S32;
typedef     int16_t  S16;
typedef      int8_t  S08;
typedef       float  F32;
typedef      double  F64;
typedef long double  F80;
typedef       void*  PTR;
typedef        char  C08;
typedef     wchar_t  C16;
typedef         S32  INT;
typedef         F64  FLO;
typedef         C08  CHR;
typedef        bool  BIT;
typedef const char*  STR;
typedef         int  IND; //for loop indexes

//DELTA API TYPES
typedef     int64_t  NUM;
typedef     int64_t  BLL;
typedef        void  VOI;
typedef      double  POS;


struct DGEID;
struct FILES;
struct DLTID;
struct MUSIC;
struct SOUND;
struct OBJID;
struct LAYER;
struct FTYPE;
struct DSXID;
struct TXTID;
struct PTHID;
struct LUCID;

struct ByteCode
{
    //DEV: should hold a small index offset array and then data bytes.
    //the byte structure must make it past the EXE/DLL boundary.
    //DESIGN: Generally only up to 8 arguments are allowed in the index array.
    //DESIGN: The bytecode is passed to the EXE as a pointer to the struct.
    //        This struct may be a temporary object but should exist for the duration of the call.
    unsigned char data_[4096];
    
    ByteCode();
    ByteCode(NUM a, NUM b);
    
    operator TXT ();
    operator NUM ();
    operator FLO ();
    operator Vx<NUM> ();
    operator Vx<FLO> ();
    operator Vx<DSXID> ();
    operator Vx<OBJID> ();
    operator DGEID ();
    operator FILES ();
    operator DLTID ();
    operator MUSIC ();
    operator SOUND ();
    operator OBJID ();
    operator LAYER ();
    operator FTYPE ();
    operator DSXID ();
    operator TXTID ();
    operator PTHID ();
    operator LUCID ();
    
    //ByteCode Access Functions
    NUM i(NUM index);
    FLO f(NUM index);
    const char* t(NUM index);
    Vx<NUM> ii(NUM index);
    Vx<FLO> ff(NUM index);
    Vx<DSXID> dd(NUM index);
    Vx<OBJID> oo(NUM index);
    
    //ByteCode Return Functions
    ByteCode(TXT data);
    ByteCode(NUM data);
    ByteCode(FLO data);
    ByteCode(Vx<NUM> data);
    ByteCode(Vx<FLO> data);
    ByteCode(DGEID data);
    ByteCode(FILES data);
    ByteCode(DLTID data);
    ByteCode(MUSIC data);
    ByteCode(SOUND data);
    ByteCode(OBJID data);
    ByteCode(LAYER data);
    ByteCode(FTYPE data);
    ByteCode(DSXID data);
    ByteCode(TXTID data);
    ByteCode(PTHID data);
    ByteCode(LUCID data);
    ByteCode(Vx<OBJID> da);
    
    //ByteCode Builder Functions
    ByteCode& operator()();
    ByteCode& operator()(int data);
    ByteCode& operator()(NUM a, NUM b);
    ByteCode& operator()(TXT data);
    ByteCode& operator()(NUM data);
    ByteCode& operator()(FLO data);
    ByteCode& operator()(Vx<NUM> data);
    ByteCode& operator()(Vx<FLO> data);
    ByteCode& operator()(Vx<OBJID> da);
    ByteCode& operator()(DGEID data);
    ByteCode& operator()(FILES data);
    ByteCode& operator()(DLTID data);
    ByteCode& operator()(MUSIC data);
    ByteCode& operator()(SOUND data);
    ByteCode& operator()(OBJID data);
    ByteCode& operator()(LAYER data);
    ByteCode& operator()(FTYPE data);
    ByteCode& operator()(DSXID data);
    ByteCode& operator()(TXTID data);
    ByteCode& operator()(PTHID data);
    ByteCode& operator()(LUCID data);
    ByteCode& operator()(ByteCode& r);
    U32& Table(NUM index);
    U32 Set(U32 where, TXT text);
    U32 Set(U32 where, NUM var);
    U32 Set(U32 where, FLO var);
    U32 Set(U32 where, Vx<NUM> var);
    U32 Set(U32 where, Vx<FLO> var);
    U32 Set(U32 where, Vx<OBJID> var);
    NUM GetA();
    NUM GetB();
    
};

struct DGEID {
    U32 v;
    operator U32(){return v;}
    DGEID(U32 vv){v = vv;}
    DGEID(){v = 0xFFFFFFFF;}
    DGEID(ByteCode data){v = data.i(0);}
};
struct FILES {
    U32 v;
    operator U32(){return v;}
    FILES(U32 vv){v = vv;}
    FILES(){v = 0xFFFFFFFF;}
    FILES(ByteCode data){v = data.i(0);}
};
struct DLTID {
    U32 v;
    operator U32(){return v;}
    DLTID(U32 vv){v = vv;}
    DLTID(){v = 0xFFFFFFFF;}
    DLTID(ByteCode data){v = data.i(0);}
};
struct MUSIC {
    U32 v;
    operator U32(){return v;}
    MUSIC(U32 vv){v = vv;}
    MUSIC(){v = 0xFFFFFFFF;}
    MUSIC(ByteCode data){v = data.i(0);}
};
struct SOUND {
    U32 v;
    operator U32(){return v;}
    SOUND(U32 vv){v = vv;}
    SOUND(){v = 0xFFFFFFFF;}
    SOUND(ByteCode data){v = data.i(0);}
};
struct OBJID {
    U32 v;
    operator U32(){return v;}
    OBJID(U32 vv){v = vv;}
    OBJID(){v = 0xFFFFFFFF;}
    OBJID(ByteCode data){v = data.i(0);}
};
struct LAYER {
    U32 v;
    operator U32(){return v;}
    LAYER(U32 vv){v = vv;}
    LAYER(){v = 0xFFFFFFFF;}
    LAYER(ByteCode data){v = data.i(0);}
};
struct FTYPE {
    U32 v;
    operator U32(){return v;}
    FTYPE(U32 vv){v = vv;}
    FTYPE(){v = 0xFFFFFFFF;}
    FTYPE(ByteCode data){v = data.i(0);}
};
struct DSXID {
    U32 v;
    operator U32(){return v;}
    DSXID(U32 vv){v = vv;}
    DSXID(){v = 0xFFFFFFFF;}
    DSXID(ByteCode data){v = data.i(0);}
};
struct TXTID {
    U32 v;
    operator U32(){return v;}
    TXTID(U32 vv){v = vv;}
    TXTID(){v = 0xFFFFFFFF;}
    TXTID(ByteCode data){v = data.i(0);}
};
struct PTHID {
    U32 v;
    operator U32(){return v;}
    PTHID(U32 vv){v = vv;}
    PTHID(){v = 0xFFFFFFFF;}
    PTHID(ByteCode data){v = data.i(0);}
};

//This type doesn't carry the same methods in the DLL that are in the EXE.
//Also, DELTA internally stores LUCIDs as raw bytes, but object DLLs use text strings.
struct LUCID {
    TXT v;
    operator TXT(){return v;}
    LUCID(TXT vv){v = vv;}
    LUCID(){v = "00000000000000000000000000000000";}
    LUCID(ByteCode data){v = data.t(0);}
};
